"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import email.parser
import io
import logging
import os
import re

import vcsinfo


LOGGER = logging.getLogger(__name__)


class VCSPyInfo(vcsinfo.VCS):
    """Minimal class for querying Python info files"""

    VERSION_RE = re.compile(r'(?P<coarse>[.\d]*\d)\.(?P<number>\d+)(.dev(?P<modified>\d+))?')

    def __init__(self, dirname):
        vcsinfo.VCS.__init__(self)

        # Look for a specific .egg-info subdirectory which will contain more information
        # than just a PKG-INFO in dirname.
        files = os.listdir(dirname)
        contents = [ent for ent in files if ent.endswith('.egg-info')]
        if contents:
            LOGGER.debug(f'Found {contents}')
            _dirname = os.path.join(dirname, contents[0])
            LOGGER.debug(f'Picking {_dirname}')
        else:
            # Didn't find a .egg-info - stick with dirname
            _dirname = dirname
        self.detect_source_root(_dirname)
        LOGGER.debug(f'Matched {self.vcs}: {dirname}')


    def detect_source_root(self, dirname):
        '''Identify if a directory might be an .info-info directory.'''

        dirname = os.path.realpath(dirname)
        try:
            pi_path = os.path.join(dirname, 'PKG-INFO')
            LOGGER.debug(f'Reading {pi_path}')
            with io.open(pi_path, encoding='utf-8', errors='replace') as piobj:
                raw_pi = piobj.read()
            self.pkg_info = email.parser.Parser().parsestr(raw_pi)
        except (ValueError, IOError) as err:
            raise TypeError(f'Directory "{dirname}" is not an pkg_info directory: {err}') from err
        self.source_root = dirname
        sources_txt = os.path.join(dirname, 'SOURCES.txt')
        try:
            with io.open(sources_txt, encoding='utf-8', errors='replace') as st_obj:
                self.files = [line.rstrip(os.linesep) for line in st_obj.readlines()]
        except IOError as err:
            raise TypeError(f'Failed reading {sources_txt}: {err}') from err


    def _get_version(self):
        return self.pkg_info.get('Version')


    def _match_version(self):
        try:
            mobj = self._ver_mobj
        except AttributeError:
            # pylint: disable=attribute-defined-outside-init
            mobj = self._ver_mobj = re.match(self.VERSION_RE, self._get_version())
        return mobj


    @property
    def name(self):
        return self.pkg_info.get('Name')


    @property
    def branch(self):
        mobj = self._match_version()
        if mobj:
            branch = mobj.groupdict().get('coarse', '0') or '0'
        else:
            branch = '0'
        return branch


    @property
    def user(self):
        return None


    @property
    def id(self):
        mobj = self._match_version()
        if mobj:
            ids = '.'.join((
                mobj.groupdict().get('coarse', '0') or '0',
                mobj.groupdict().get('number', '0') or '0',
            ))
        else:
            ids = '0'
        return ids


    @property
    def id_short(self):
        return self.id


    @property
    def number(self):
        mobj = self._match_version()
        if mobj:
            number = mobj.groupdict().get('number', '0') or '0'
        else:
            number = '0'
        return int(number)


    def list_files(self):
        return self.files


    @property
    def modified(self):
        mobj = re.match(self.VERSION_RE, self.id)
        if mobj:
            mod = mobj.groupdict().get('modified', '0') or '0'
        else:
            mod = '0'
        return int(mod)


VCS = VCSPyInfo


# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import shutil
import tempfile

import pytest

from vcsinfo import detect_vcs


# Since we cannot have a git repo checked into another repo, instead use a git repo that was previously
# created and the .git folder renamed and then move it to another location and rename the .git folder
REPO_DIR = os.path.join(os.path.dirname(__file__), 'repos/git-repo')


@pytest.fixture(name='git_repo_path')
def fixture_git_repo_path():
    with tempfile.TemporaryDirectory(suffix='-test-vcsinfo-git') as temp_dir:
        git_repo_path = os.path.join(temp_dir, 'git-repo')
        shutil.copytree(REPO_DIR, git_repo_path)
        shutil.move(os.path.join(git_repo_path, 'dotgit'), os.path.join(git_repo_path, '.git'))
        yield git_repo_path


def test_git_repo(git_repo_path):
    result = detect_vcs(git_repo_path)
    assert result
    assert result.vcs == 'git'
    assert str(result) == f'VCSGit({git_repo_path})'
    assert result.id == 'a23bf7e941d6edf95ec1ba3dc2e4999a544c6b5d'
    # Comes from the origin repo
    assert result.name == 'vcsinfo-test-git'
    assert result.branch == 'main'
    assert result.id_short == 'a23bf7'
    assert result.id_string == f'main-2.Ia23bf7.M{result.modified}'
    assert result.modified != 0
    assert result.number == 2
    assert result.release == f'2.Ia23bf7.M{result.modified}'
    assert result.source_root == git_repo_path
    assert result.upstream_repo == 'git@github.com:adobe/vcsinfo-test-git-upstream.git'
    assert result.user == 'saville'
    assert result.list_files() == ['hello', 'hello2', 'hello3']
    assert result.info() == {
        'type': 'git',
        'upstream_repo': 'git@github.com:adobe/vcsinfo-test-git-upstream.git',
        'name': 'vcsinfo-test-git',
        'branch': 'main',
        'id': 'a23bf7e941d6edf95ec1ba3dc2e4999a544c6b5d',
        'id_short': 'a23bf7',
        'id_string': f'main-2.Ia23bf7.M{result.modified}',
        'number': 2,
        'user': 'saville',
        'release': f'2.Ia23bf7.M{result.modified}',
    }
    assert result.info(include_files=True)['files'] == ['hello', 'hello2', 'hello3']
    assert result.status() == [
        ['hello'], ['hello3'], [], [], [], [], ['hello2']
    ]



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import glob
import logging
import os
from typing import List


LOGGER = logging.getLogger(__file__)


class VCSUnsupported(Exception):
    """Custom error signifying a specific VCS is not supported"""


def search_parent_dirs(directory, filename):
    """
    Search a directory and each parent directory for a filename.
    If found return the full path; otherwise return ''.
    """

    found = ''

    dir_parts = directory.split(os.path.sep)
    while dir_parts:
        filename_parts = dir_parts + [filename]
        filename_path = os.path.sep.join(filename_parts)
        if os.path.exists(filename_path):
            found = filename_path
            break
        dir_parts.pop()

    return found


def last_mtime(files, prefix=None):
    """
    Find the last modified time for a given list of files.
    """
    mtime = None

    if files:
        if prefix:
            _files = [os.path.join(prefix, f) for f in files]
        else:
            _files = files

        # NOTE: If the file doesn't exist then return 0, but then make sure a
        # 0 result gets mapped back to None.
        mtime = int(max([
            os.path.exists(fn) and os.stat(fn).st_mtime or 0 for fn in _files
        ])) or None

    return mtime


ST_MOD = 0
ST_ADD = 1
ST_REM = 2
ST_DEL = 3
ST_UNK = 4
ST_IGN = 5
ST_CLN = 6


class VCS:
    """Base class for a source tree managed by VCS"""

    def __init__(self):
        """Base constructor"""
        # Automatically set the VCS engine name
        self.vcs = self.__class__.__name__
        if self.vcs.startswith('VCS'):
            self.vcs = self.vcs[3:].lower()

        self.source_root = ''
        self._name = ''
        self._upstream_repo = ''

    @property
    def upstream_repo(self):
        """The location of the up-stream VCS repository."""
        return self._upstream_repo

    @property
    def name(self):
        """The name of the VCS project."""
        return self._name

    @property
    # pylint: disable=R0201
    def branch(self):
        """The branch name."""
        return None

    @property
    # pylint: disable=R0201,invalid-name
    def id(self):
        """The native commit identification."""
        return None

    @property
    # pylint: disable=R0201
    def user(self):
        """The user who performed the commit (if applicable)."""
        return 'n/a'

    @property
    def id_short(self):
        """
        Differs from .id for VCS that have large hash values that
        can be abbreviated.
        """
        return self.id

    @property
    def number(self):
        """
        This is the same as .id if commits have increasing commit numbers.
        Otherwise this simulates an increasing commit number.
        """
        return int(self.id)

    @property
    def modified(self):
        """
        Return none if there are no local modifications to the VCS checkout,
        otherwise return epoc seconds of the most recently modified file.

        NOTE: modified will update each time a file in the changeset is
        modified.  It cannot update, however, when a REMOVED file is added
        to the changeset because there may no longer be a file in the
        file system to stat() to get the removed time!  This means that
        modified can only updated with MODIFIED and ADDED files.
        """
        return last_mtime(
            sum(self.status()[ST_MOD:ST_ADD + 1], []),
            prefix=self.source_root
        ) or 0

    @property
    def release(self):
        """
        String representing the current state of the branch, including the
        commit number, id_short, and whether any modifications have been made
        to the local tree.
        """
        modified = self.modified
        modstr = f'.M{modified}'

        rel_str = f'{self.number}.I{self.id_short}{modstr}'

        return rel_str

    @property
    def id_string(self):
        """
        String providing a unique id of the branch and release.
        """
        return str(self.branch) + "-" + str(self.release)

    # pylint: disable=R0201
    def status(self):
        """
        Returns a tuple of lists of modified files in the changeset:
        (
            [MODIFIED, ...], # ST_MOD: changeset is modifying files
            [ADDED, ...],    # ST_ADD: changeset is adding file
            [REMOVED, ...],  # ST_REM: changeset is removing file
            [DELETED, ...],  # ST_DEL: file is missing but not in changeset
            [UNKNOWN, ...],  # ST_UNK: file exists but not in changeset
            [IGNORED, ...],  # ST_IGN: file exists but VCS explicitly ignores
            [CLEAN, ...],    # ST_CLN: file tracked by VCS and unmodified
        )
        """
        return [], [], [], [], [], [], []

    def list_files(self):
        """
        Return a list of files known to the VCS
        (committed + changeset(added, moved) - changeset(deleted))
        """
        raise NotImplementedError(f'VCS module {self.vcs} must implement list_files()')

    def info(self, include_files=False):
        """
        Return vcs information as a dict, optionally including full file
        information.
        """
        info = {
            'type': self.vcs,
            'upstream_repo': self.upstream_repo,
            'name': self.name,
            'branch': self.branch,
            'id': self.id,
            'id_short': self.id_short,
            'id_string': self.id_string,
            'number': self.number,
            'user': self.user,
            'release': self.release,
        }

        if include_files:
            info['files'] = self.list_files()
            # Do not report UNKNOWN or IGNORED files
            status_info = list(self.status())
            status_info[ST_UNK] = []
            status_info[ST_IGN] = []
            info['filestatus'] = status_info

        return info

    def __str__(self):
        """String represenation"""
        return f'{self.__class__.__name__}({self.source_root})'


def load_vcs(name, directory, *args, **argv) -> VCS:
    """
    Load a specific vcs module for the given directory and arguments.

    Args:
        name: the vcs module name
        directory: the directory to load the vcs module for
    Raises:
        VCSUnsupported: if the vcs module does not support the given directory
        TypeError: if the vcs module does not support the given directory
    """
    LOGGER.debug(f'Loading VCS module: {name}')
    vcs_module = __import__('.'.join((__name__, name)), fromlist=[__name__])
    vcs = vcs_module.VCS(directory, *args, **argv)
    return vcs


def detect_vcss(directory, *args, **argv) -> List[VCS]:
    """
    Interrogate the given directory for vcs information, returning an object
    that can be used to obtain information about the current conditions of the
    source tree.
    """
    directory = directory or os.getcwd()
    vcs_dir = os.path.dirname(__file__)
    # Ignore things that don't look like python modules
    # and list archive.py last.
    vcs_files = glob.glob(os.path.join(vcs_dir, '*.py'))
    possible_vcs = []

    errors = []
    for modpath in vcs_files:
        modname, _ = os.path.splitext(os.path.basename(modpath))
        # skip things like '__init__.py'
        if modname.startswith('_') or modname == 'version':
            continue

        try:
            vcs = load_vcs(modname, directory, *args, **argv)
            if hasattr(vcs, 'is_archive') and vcs.is_archive:
                # if an "archive" vcs config is present, it always wins
                return [vcs]

            possible_vcs.append(vcs)
        except (VCSUnsupported, TypeError) as err:
            LOGGER.debug(f'Failed {modname}: {err}')
            errors.append(str(err))

    if not possible_vcs:
        # pylint: disable=C0301
        message = f'No recognized VCS management of source tree "{directory}" - do you need to login to a VCS?'
        for error in errors:
            message += f'\n\tWARNING: {error}'
        raise VCSUnsupported(message)

    return possible_vcs


def detect_vcs(directory, *args, **argv) -> VCS:
    """
    Detects the VCS type for a given directory.
    :param directory: the directory to determine the VCS for
    :return: the VCS type
    """
    possible_vcss = detect_vcss(directory, *args, **argv)
    LOGGER.debug(f'Possible VCSs: {possible_vcss}', possible_vcss)

    if len(possible_vcss) > 1:
        possible_vcss_str = ', '.join([vcs.vcs for vcs in possible_vcss])
        LOGGER.warning(f'WARNING: Multiple VCS matches: {possible_vcss_str}')

    return possible_vcss[0]


try:
    import setuptools.command.egg_info

    # pyline: disable=R0904
    class VCSInfoEggInfo(setuptools.command.egg_info.egg_info):
        """
        Override the egg_info command to appropriately set build tags.
        """

        command_name = 'egg_info'

        def tags(self):
            sourcedir = os.path.dirname(
                os.path.abspath(self.distribution.script_name)
            )

            # first, see if we can get vcs information.
            try:
                vcs = detect_vcs(sourcedir)
                tag_build = f'.{vcs.number}'
                if vcs.modified > 0:
                    tag_build = f'{tag_build}.{vcs.modified}'
                return tag_build
            except TypeError:
                pass
            except VCSUnsupported:
                pass

            # if not fall back to the default routine
            return setuptools.command.egg_info.egg_info.tags(self)


except ImportError:
    pass


# Local Variables:
# fill-column: 100
# End:



import importlib.machinery
import os
import subprocess
import types
from setuptools import setup, find_packages


BASE_VERSION = '2.1'
SOURCE_DIR = os.path.dirname(
    os.path.abspath(__file__)
)
VCSINFO_DIR = os.path.join(SOURCE_DIR, 'vcsinfo')
VERSION_FILE = os.path.join(VCSINFO_DIR, 'version.py')
HEADER_FILE = os.path.join(SOURCE_DIR, '.pylint-license-header')


with open(os.path.join(SOURCE_DIR, 'README.rst')) as fobj:
    long_description = fobj.read().strip()


def get_version():
    """
    Call out to the git command line to get the current commit "number".
    """
    if os.path.exists(VERSION_FILE):
        # Read version from file
        loader = importlib.machinery.SourceFileLoader('vcsinfo_version', VERSION_FILE)
        version_mod = types.ModuleType(loader.name)
        loader.exec_module(version_mod)
        existing_version = version_mod.__version__  # pylint: disable=no-member
        print(f'Using existing vcsinfo version: {existing_version}')
        return existing_version

    # Generate the version from the base version and the git commit number, then store it in the file
    try:
        cmd = subprocess.Popen(
            args=[
                'git',
                'rev-list',
                '--count',
                'HEAD',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding='utf8',
        )
        stdout = cmd.communicate()[0]
        output = stdout.strip()
        if cmd.returncode == 0:
            new_version = '{0}.{1}'.format(BASE_VERSION, output)
            print(f'Setting version to {new_version}')

            # write the version file
            if os.path.exists(VCSINFO_DIR):
                with open(HEADER_FILE, 'r', encoding='utf8') as fobj:
                    header = fobj.read()
                with open(VERSION_FILE, 'w', encoding='utf8') as fobj:
                    fobj.write(f"{header}\n__version__ = '{new_version}'\n")
            return new_version
    except Exception as exc:
        print(f'Could not generate version from git commits: {exc}')
    # If all else fails, use development version
    return f'{BASE_VERSION}.DEVELOPMENT'


# pylint: disable=C0301
setup(
    name='vcsinfo',
    version=get_version(),
    author='Adobe',
    author_email='noreply@adobe.com',
    license='MIT',
    url='https://github.com/adobe/vcsinfo',
    description='Utilities to normalize working with different Version Control Systems',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=("tests",)),
    scripts=[
        'bin/vcsinfo',
    ],
    install_requires=[
        'GitPython<4',
        'gitdb<5',
    ],
    extras_require={
        # Make everything except git an optional dependency
        'hg': ['mercurial'],
        'p4': ['p4python'],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""
__version__ = '2.1.103'
